searchData={"content_type":"text/markdown","items":[{"doc":"A libcluster strategy that uses Postgres LISTEN/NOTIFY to determine the cluster topology.\n\nThis strategy operates by having all nodes in the cluster listen for and send notifications to a shared Postgres channel.\n\nWhen a node comes online, it begins to broadcast its name in a \"heartbeat\" message to the channel. All other nodes that receive this message attempt to connect to it.\n\nThis strategy does not check connectivity between nodes and does not disconnect them","ref":"LibclusterPostgres.Strategy.html","title":"LibclusterPostgres.Strategy","type":"module"},{"doc":"* `heartbeat_interval` - The interval at which to send heartbeat messages in milliseconds (optional; default: 5_000)\n* `channel_name` - The name of the channel to which nodes will listen and notify (optional; defaults to the result of `Node.get_cookie/0`)","ref":"LibclusterPostgres.Strategy.html#module-options","title":"Options - LibclusterPostgres.Strategy","type":"module"},{"doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"LibclusterPostgres.Strategy.html#child_spec/1","title":"LibclusterPostgres.Strategy.child_spec/1","type":"function"},{"doc":"","ref":"LibclusterPostgres.Strategy.html#handle_continue/2","title":"LibclusterPostgres.Strategy.handle_continue/2","type":"function"},{"doc":"","ref":"LibclusterPostgres.Strategy.html#init/1","title":"LibclusterPostgres.Strategy.init/1","type":"function"},{"doc":"","ref":"LibclusterPostgres.Strategy.html#start_link/1","title":"LibclusterPostgres.Strategy.start_link/1","type":"function"},{"doc":"# libcluster Postgres Strategy\n\n[![Hex version badge](https://img.shields.io/hexpm/v/libcluster_postgres.svg)](https://hex.pm/packages/libcluster_postgres)\n[![License badge](https://img.shields.io/hexpm/l/libcluster_postgres.svg)](https://github.com/supabase/libcluster_postgres/blob/main/LICENSE)\n[![Elixir CI](https://github.com/supabase/libcluster_postgres/actions/workflows/elixir.yaml/badge.svg)](https://github.com/supabase/libcluster_postgres/actions/workflows/elixir.yaml)\n\nPostgres Strategy for [libcluster](https://hexdocs.pm/libcluster/) which is used by Supabase on the [realtime](https://github.com/supabase/realtime), [supavisor](https://github.com/supabase/supavisor) and [logflare](https://github.com/logflare/logflare) projects.\n\nYou can test it out by running `docker compose up`\n\n![example.png](example.png)","ref":"readme.html","title":"libcluster Postgres Strategy","type":"extras"},{"doc":"The package can be installed\nby adding `libcluster_postgres` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [{:libcluster_postgres, \"~> 0.1\"}]\nend\n```","ref":"readme.html#installation","title":"Installation - libcluster Postgres Strategy","type":"extras"},{"doc":"To use it, set your configuration file with the informations for your database:\n\n```elixir\nconfig :libcluster,\n  topologies: [\n    example: [\n      strategy: Cluster.Strategy.Postgres,\n      config: [\n          hostname: \"localhost\",\n          username: \"postgres\",\n          password: \"postgres\",\n          database: \"postgres\",\n          port: 5432,\n          parameters: [],\n          # optional, defaults to node cookie\n          channel_name: \"cluster\"\n      ],\n    ]\n  ]\n```\n\nThen add it to your supervision tree:\n\n```elixir\ndefmodule MyApp do\n  use Application\n\n  def start(_type, _args) do\n    topologies = Application.get_env(:libcluster, :topologies)\n    children = [\n      # ...\n      {Cluster.Supervisor, [topologies]}\n      # ...\n    ]\n    Supervisor.start_link(children, strategy: :one_for_one)\n  end\nend\n```\n#","ref":"readme.html#how-to-use-it","title":"How to use it - libcluster Postgres Strategy","type":"extras"},{"doc":"At Supabase, we use clustering in all of our Elixir projects which include [Logflare](https://github.com/Logflare/logflare), [Supavisor](https://github.com/supabase/supavisor) and [Realtime](https://github.com/supabase/realtime). With multiple servers connected we can load shed, create globally distributed services and provide the best service to our customers so we’re closer to them geographically and to their instances, reducing overall latency.\n\n![Example of Realtime architecture where a customer from CA will connect to the server closest to them and their Supabase instance](realtime_example.png)\n\nExample of Realtime architecture where a customer from CA will connect to the server closest to them and their Supabase instance\nTo achieve a connected cluster, we wanted to be as cloud-agnostic as possible. This makes our self-hosting options more accessible. We don’t want to introduce extra services to solve this single issue - Postgres is the logical way to achieve it.\n\nThe other piece of the puzzle was already built by the Erlang community being the defacto library to facilitate the creation of connected Elixir servers: [libcluster](https://github.com/bitwalker/libcluster).\n\n#","ref":"readme.html#why-do-we-need-a-distributed-erlang-cluster","title":"Why do we need a distributed Erlang Cluster? - libcluster Postgres Strategy","type":"extras"},{"doc":"[libcluster](https://github.com/bitwalker/libcluster) is the go-to package for connecting multiple BEAM instances and setting up healing strategies. libcluster provides out-of-the-box strategies and it allows users to define their own strategies by implementing a simple behavior that defines cluster formation and healing according to the supporting service you want to use.\n\n#","ref":"readme.html#what-is-libcluster","title":"What is libcluster? - libcluster Postgres Strategy","type":"extras"},{"doc":"Postgres provides an event system using two commands: [NOTIFY](https://www.postgresql.org/docs/current/sql-notify.html) and [LISTEN](https://www.postgresql.org/docs/current/sql-listen.html) so we can use them to propagate events within our Postgres instance.\n\nTo use this features, you can use psql itself or any other Postgres client. Start by listening on a specific channel, and then notify to receive a payload.\n\n```markdown\npostgres=# LISTEN channel;\nLISTEN\npostgres=# NOTIFY channel, 'payload';\nNOTIFY\nAsynchronous notification \"channel\" with payload \"payload\" received from server process with PID 326.\n```\n\nNow we can replicate the same behavior in Elixir and [Postgrex](https://hex.pm/packages/postgrex) within IEx (Elixir's interactive shell).\n\n```elixir\nMix.install([{:postgrex, \"~> 0.17.3\"}])\nconfig = [\n  hostname: \"localhost\",\n  username: \"postgres\",\n  password: \"postgres\",\n  database: \"postgres\",\n  port: 5432\n]\n{:ok, db_notification_pid} = Postgrex.Notifications.start_link(config)\nPostgrex.Notifications.listen!(db_notification_pid, \"channel\")\n{:ok, db_conn_pid} = Postgrex.start_link(config)\nPostgrex.query!(db_conn_pid, \"NOTIFY channel, 'payload'\", [])\n\nreceive do msg -> IO.inspect(msg) end\n# Mailbox will have a message with the following content:\n# {:notification, #PID , #Reference , \"channel\", \"test\"}\n```\n\n#","ref":"readme.html#how-did-we-use-postgres","title":"How did we use Postgres? - libcluster Postgres Strategy","type":"extras"},{"doc":"Using the libcluster `Strategy` behavior, inspired by [this GitHub repository](https://github.com/kevbuchanan/libcluster_postgres) and knowing how `NOTIFY/LISTEN` works, implementing a strategy becomes straightforward:\n\n1. We send a `NOTIFY` to a channel with our `node()` address to a configured channel\n\n```elixir\n# lib/cluster/strategy/postgres.ex:52\ndef handle_continue(:connect, state) do\n    with {:ok, conn}  \n        Logger.error(state.topology, \"Failed to connect to Postgres: #{inspect(reason)}\")\n        {:noreply, state}\n    end\n  end\n```\n\n2. We actively listen for new `{:notification, pid, reference, channel, payload}` messages and connect to the node received in the payload\n\n```elixir\n# lib/cluster/strategy/postgres.ex:80\ndef handle_info({:notification, _, _, _, node}, state) do\n    node = String.to_atom(node)\n\n    if node != node() do\n      topology = state.topology\n      Logger.debug(topology, \"Trying to connect to node: #{node}\")\n\n      case Strategy.connect_nodes(topology, state.connect, state.list_nodes, [node]) do\n        :ok -> Logger.debug(topology, \"Connected to node: #{node}\")\n        {:error, _} -> Logger.error(topology, \"Failed to connect to node: #{node}\")\n      end\n    end\n\n    {:noreply, state}\n  end\n```\n\n3. Finally, we configure a heartbeat that is similar to the first message sent for cluster formation so libcluster is capable of heal if need be\n\n```elixir\n# lib/cluster/strategy/postgres.ex:73\ndef handle_info(:heartbeat, state) do\n    Process.cancel_timer(state.meta.heartbeat_ref)\n    Postgrex.query(state.meta.conn, \"NOTIFY #{state.config[:channel_name]}, '#{node()}'\", [])\n    ref = heartbeat(state.config[:heartbeat_interval])\n    {:noreply, put_in(state.meta.heartbeat_ref, ref)}\nend\n```\n\nThese three simple steps allow us to connect as many nodes as needed, regardless of the cloud provider, by utilising something that most projects already have: a Postgres connection.","ref":"readme.html#building-the-strategy","title":"Building the strategy - libcluster Postgres Strategy","type":"extras"},{"doc":"A special thank you to [@gotbones](https://twitter.com/gotbones) for creating libcluster and [@kevinbuch\\_](https://twitter.com/kevinbuch_) for the original inspiration for this strategy.","ref":"readme.html#acknowledgements","title":"Acknowledgements - libcluster Postgres Strategy","type":"extras"}]}